@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org/projects/xpce/
-    Copyright (c)  2006-2016, University of Amsterdam
+    Copyright (c)  2006-2017, University of Amsterdam
                               VU University Amsterdam
     All rights reserved.
 
@@ -67,9 +67,10 @@
             xref_defined_class/3        % ?Source, ?ClassName, -How
           ]).
 :- use_module(library(debug), [debug/3]).
-:- use_module(library(lists), [append/3, member/2, select/3]).
+:- use_module(library(lists), [append/3, append/2, member/2, select/3]).
 :- use_module(library(operators), [push_op/3]).
 :- use_module(library(shlib), [current_foreign_library/2]).
+:- use_module(library(ordsets)).
 :- use_module(library(prolog_source)).
 :- use_module(library(option)).
 :- use_module(library(error)).
@@ -411,24 +412,36 @@ xref_input_stream(Stream) :-
 %   them to be undone by pop_operators/0.
 
 xref_push_op(Src, P, T, N0) :-
-    !,
-    (   N0 = _:_
-    ->  N = N0
-    ;   '$current_source_module'(M),
-        N = M:N0
-    ),
-    valid_op(op(P,T,N)),
-    push_op(P, T, N),
-    assert_op(Src, op(P,T,N)),
-    debug(xref(op), ':- ~w.', [op(P,T,N)]).
+    '$current_source_module'(M0),
+    strip_module(M0:N0, M, N),
+    (   is_list(N),
+        N \== []
+    ->  maplist(push_op(Src, P, T, M), N)
+    ;   push_op(Src, P, T, M, N)
+    ).
 
-valid_op(op(P,T,N)) :-
-    atom(N),
+push_op(Src, P, T, M0, N0) :-
+    strip_module(M0:N0, M, N),
+    Name = M:N,
+    valid_op(op(P,T,Name)),
+    push_op(P, T, Name),
+    assert_op(Src, op(P,T,Name)),
+    debug(xref(op), ':- ~w.', [op(P,T,Name)]).
+
+valid_op(op(P,T,M:N)) :-
+    atom(M),
+    valid_op_name(N),
     integer(P),
     between(0, 1200, P),
     atom(T),
     op_type(T).
 
+valid_op_name(N) :-
+    atom(N),
+    !.
+valid_op_name(N) :-
+    N == [].
+
 op_type(xf).
 op_type(yf).
 op_type(fx).
@@ -928,6 +941,8 @@ process_directive(public(Public), Src) :-
     process_predicates(assert_public, Public, Src).
 process_directive(export(Export), Src) :-
     process_predicates(assert_export, Export, Src).
+process_directive(import(Import), Src) :-
+    process_import(Import, Src).
 process_directive(module(Module, Export), Src) :-
     assert_module(Src, Module),
     assert_module_export(Src, Export).
@@ -1142,6 +1157,7 @@ xref_meta(call_dcg(G, _A, _R),  [//(G)]).
 xref_meta(phrase_from_file(G,_),[//(G)]).
 xref_meta(catch(A, _, B),       [A, B]).
 xref_meta(thread_create(A,_,_), [A]).
+xref_meta(thread_create(A,_),   [A]).
 xref_meta(thread_signal(_,A),   [A]).
 xref_meta(thread_at_exit(A),    [A]).
 xref_meta(thread_initialization(A), [A]).
@@ -1296,29 +1312,31 @@ arith_callable(Name/Arity, Goal) :-
 %
 %   We limit the number of explored paths   to  100 to avoid getting
 %   trapped in this analysis.
-%
-%   @bug  We  should  analyse  whether    bindings  due  to  partial
-%   evaluation lead to a different analysis.
 
 process_body(Body, Origin, Src) :-
-    forall(limit(100, process_goal(Body, Origin, Src)),
+    forall(limit(100, process_goal(Body, Origin, Src, _Partial)),
            true).
 
-process_goal(Var, _, _) :-
+%!  process_goal(+Goal, +Origin, +Src, ?Partial) is multi.
+%
+%   Xref Goal. The argument Partial is bound   to  `true` if there was a
+%   partial evalation inside Goal that has bound variables.
+
+process_goal(Var, _, _, _) :-
     var(Var),
     !.
-process_goal(Goal, Origin, Src) :-
-    Goal = (_;_),
+process_goal(Goal, Origin, Src, P) :-
+    Goal = (_,_),                               % problems
+    !,
+    phrase(conjunction(Goal), Goals),
+    process_conjunction(Goals, Origin, Src, P).
+process_goal(Goal, Origin, Src, _) :-           % Final disjunction, no
+    Goal = (_;_),                               % problems
     !,
     phrase(disjunction(Goal), Goals),
-    setof(Goal,
-          (   member(G, Goals),
-              process_goal(G, Origin, Src)
-          ),
-          Alts0),
-    variants(Alts0, 10, Alts),
-    member(Goal, Alts).
-process_goal(Goal, Origin, Src) :-
+    forall(member(G, Goals),
+           process_body(G, Origin, Src)).
+process_goal(Goal, Origin, Src, P) :-
     (   (   xmodule(M, Src)
         ->  true
         ;   M = user
@@ -1333,95 +1351,145 @@ process_goal(Goal, Origin, Src) :-
     !,
     must_be(list, Called),
     assert_called(Src, Origin, Goal),
-    process_called_list(Called, Origin, Src).
-process_goal(Goal, Origin, Src) :-
+    process_called_list(Called, Origin, Src, P).
+process_goal(Goal, Origin, Src, _) :-
     process_xpce_goal(Goal, Origin, Src),
     !.
-process_goal(load_foreign_library(File), _Origin, Src) :-
+process_goal(load_foreign_library(File), _Origin, Src, _) :-
     process_foreign(File, Src).
-process_goal(load_foreign_library(File, _Init), _Origin, Src) :-
+process_goal(load_foreign_library(File, _Init), _Origin, Src, _) :-
     process_foreign(File, Src).
-process_goal(use_foreign_library(File), _Origin, Src) :-
+process_goal(use_foreign_library(File), _Origin, Src, _) :-
     process_foreign(File, Src).
-process_goal(use_foreign_library(File, _Init), _Origin, Src) :-
+process_goal(use_foreign_library(File, _Init), _Origin, Src, _) :-
     process_foreign(File, Src).
-process_goal(Goal, Origin, Src) :-
+process_goal(Goal, Origin, Src, P) :-
     xref_meta_src(Goal, Metas, Src),
     !,
     assert_called(Src, Origin, Goal),
-    process_called_list(Metas, Origin, Src).
-process_goal(Goal, Origin, Src) :-
+    process_called_list(Metas, Origin, Src, P).
+process_goal(Goal, Origin, Src, _) :-
     asserting_goal(Goal, Rule),
     !,
     assert_called(Src, Origin, Goal),
     process_assert(Rule, Origin, Src).
-process_goal(Goal, Origin, Src) :-
-    partial_evaluate(Goal),
+process_goal(Goal, Origin, Src, P) :-
+    partial_evaluate(Goal, P),
     assert_called(Src, Origin, Goal).
 
 disjunction(Var)   --> {var(Var), !}, [Var].
 disjunction((A;B)) --> !, disjunction(A), disjunction(B).
 disjunction(G)     --> [G].
 
-process_called_list([], _, _).
-process_called_list([H|T], Origin, Src) :-
-    process_meta(H, Origin, Src),
-    process_called_list(T, Origin, Src).
+conjunction(Var)   --> {var(Var), !}, [Var].
+conjunction((A,B)) --> !, conjunction(A), conjunction(B).
+conjunction(G)     --> [G].
 
-process_meta(A+N, Origin, Src) :-
+shares_vars(RVars, T) :-
+    term_variables(T, TVars0),
+    sort(TVars0, TVars),
+    ord_intersect(RVars, TVars).
+
+process_conjunction([], _, _, _).
+process_conjunction([Disj|Rest], Origin, Src, P) :-
+    nonvar(Disj),
+    Disj = (_;_),
+    Rest \== [],
+    !,
+    phrase(disjunction(Disj), Goals),
+    term_variables(Rest, RVars0),
+    sort(RVars0, RVars),
+    partition(shares_vars(RVars), Goals, Sharing, NonSHaring),
+    forall(member(G, NonSHaring),
+           process_body(G, Origin, Src)),
+    (   Sharing == []
+    ->  true
+    ;   maplist(term_variables, Sharing, GVars0),
+        append(GVars0, GVars1),
+        sort(GVars1, GVars),
+        ord_intersection(GVars, RVars, SVars),
+        VT =.. [v|SVars],
+        findall(VT,
+                (   member(G, Sharing),
+                    process_goal(G, Origin, Src, PS),
+                    PS == true
+                ),
+                Alts0),
+        (   Alts0 == []
+        ->  true
+        ;   (   true
+            ;   P = true,
+                sort(Alts0, Alts1),
+                variants(Alts1, 10, Alts),
+                member(VT, Alts)
+            )
+        )
+    ),
+    process_conjunction(Rest, Origin, Src, P).
+process_conjunction([H|T], Origin, Src, P) :-
+    process_goal(H, Origin, Src, P),
+    process_conjunction(T, Origin, Src, P).
+
+
+process_called_list([], _, _, _).
+process_called_list([H|T], Origin, Src, P) :-
+    process_meta(H, Origin, Src, P),
+    process_called_list(T, Origin, Src, P).
+
+process_meta(A+N, Origin, Src, P) :-
     !,
     (   extend(A, N, AX)
-    ->  process_goal(AX, Origin, Src)
+    ->  process_goal(AX, Origin, Src, P)
     ;   true
     ).
-process_meta(//(A), Origin, Src) :-
+process_meta(//(A), Origin, Src, P) :-
     !,
-    process_dcg_goal(A, Origin, Src).
-process_meta(G, Origin, Src) :-
-    process_goal(G, Origin, Src).
+    process_dcg_goal(A, Origin, Src, P).
+process_meta(G, Origin, Src, P) :-
+    process_goal(G, Origin, Src, P).
 
-%!  process_dcg_goal(+Grammar, +Origin, +Src) is det.
+%!  process_dcg_goal(+Grammar, +Origin, +Src, ?Partial) is det.
 %
 %   Process  meta-arguments  that  are  tagged   with  //,  such  as
 %   phrase/3.
 
-process_dcg_goal(Var, _, _) :-
+process_dcg_goal(Var, _, _, _) :-
     var(Var),
     !.
-process_dcg_goal((A,B), Origin, Src) :-
+process_dcg_goal((A,B), Origin, Src, P) :-
     !,
-    process_dcg_goal(A, Origin, Src),
-    process_dcg_goal(B, Origin, Src).
-process_dcg_goal((A;B), Origin, Src) :-
+    process_dcg_goal(A, Origin, Src, P),
+    process_dcg_goal(B, Origin, Src, P).
+process_dcg_goal((A;B), Origin, Src, P) :-
     !,
-    process_dcg_goal(A, Origin, Src),
-    process_dcg_goal(B, Origin, Src).
-process_dcg_goal((A|B), Origin, Src) :-
+    process_dcg_goal(A, Origin, Src, P),
+    process_dcg_goal(B, Origin, Src, P).
+process_dcg_goal((A|B), Origin, Src, P) :-
     !,
-    process_dcg_goal(A, Origin, Src),
-    process_dcg_goal(B, Origin, Src).
-process_dcg_goal((A->B), Origin, Src) :-
+    process_dcg_goal(A, Origin, Src, P),
+    process_dcg_goal(B, Origin, Src, P).
+process_dcg_goal((A->B), Origin, Src, P) :-
     !,
-    process_dcg_goal(A, Origin, Src),
-    process_dcg_goal(B, Origin, Src).
-process_dcg_goal((A*->B), Origin, Src) :-
+    process_dcg_goal(A, Origin, Src, P),
+    process_dcg_goal(B, Origin, Src, P).
+process_dcg_goal((A*->B), Origin, Src, P) :-
     !,
-    process_dcg_goal(A, Origin, Src),
-    process_dcg_goal(B, Origin, Src).
-process_dcg_goal({Goal}, Origin, Src) :-
+    process_dcg_goal(A, Origin, Src, P),
+    process_dcg_goal(B, Origin, Src, P).
+process_dcg_goal({Goal}, Origin, Src, P) :-
     !,
-    process_goal(Goal, Origin, Src).
-process_dcg_goal(List, _Origin, _Src) :-
+    process_goal(Goal, Origin, Src, P).
+process_dcg_goal(List, _Origin, _Src, _) :-
     is_list(List),
     !.               % terminal
-process_dcg_goal(List, _Origin, _Src) :-
+process_dcg_goal(List, _Origin, _Src, _) :-
     string(List),
     !.                % terminal
-process_dcg_goal(Callable, Origin, Src) :-
+process_dcg_goal(Callable, Origin, Src, P) :-
     extend(Callable, 2, Goal),
     !,
-    process_goal(Goal, Origin, Src).
-process_dcg_goal(_, _, _).
+    process_goal(Goal, Origin, Src, P).
+process_dcg_goal(_, _, _, _).
 
 
 extend(Var, _, _) :-
@@ -1470,7 +1538,7 @@ variants([H|T], V, Max, List) :-
         variants(T, H, Max1, List2)
     ).
 
-%!  partial_evaluate(Goal) is det.
+%!  partial_evaluate(+Goal, ?Parrial) is det.
 %
 %   Perform partial evaluation on Goal to trap cases such as below.
 %
@@ -1482,10 +1550,11 @@ variants([H|T], V, Max, List) :-
 %   @tbd    Make this user extensible? What about non-deterministic
 %           bindings?
 
-partial_evaluate(Goal) :-
+partial_evaluate(Goal, P) :-
     eval(Goal),
-    !.
-partial_evaluate(_).
+    !,
+    P = true.
+partial_evaluate(_, _).
 
 eval(X = Y) :-
     unify_with_occurs_check(X, Y).
@@ -2189,6 +2258,22 @@ assert_reexport(false, _, _) :- !.
 assert_reexport(true, Src, Term) :-
     assert(exported(Term, Src)).
 
+%!  process_import(:Import, +Src)
+%
+%   Process an import/1 directive
+
+process_import(M:PI, Src) :-
+    pi_to_head(PI, Head),
+    !,
+    (   atom(M),
+        current_module(M),
+        module_property(M, file(From))
+    ->  true
+    ;   From = '<unknown>'
+    ),
+    assert(imported(Head, Src, From)).
+process_import(_, _).
+
 %!  assert_xmodule_callable(PIs, Module, Src, From)
 %
 %   We can call all exports  and   public  predicates of an imported
@@ -2209,12 +2294,14 @@ assert_xmodule_callable([PI|T], M, Src, From) :-
 %
 %   @param Op       Ground term op(Priority, Type, Name).
 
-assert_op(Src, op(P,T,_:N)) :-
-    (   xop(Src, op(P,T,N))
+assert_op(Src, op(P,T,M:N)) :-
+    (   '$current_source_module'(M)
+    ->  Name = N
+    ;   Name = M:N
+    ),
+    (   xop(Src, op(P,T,Name))
     ->  true
-    ;   valid_op(op(P,T,N))
-    ->  assert(xop(Src, op(P,T,N)))
-    ;   true
+    ;   assert(xop(Src, op(P,T,Name)))
     ).
 
 %!  assert_module(+Src, +Module)
@@ -2227,7 +2314,12 @@ assert_module(Src, Module) :-
     !.
 assert_module(Src, Module) :-
     '$set_source_module'(Module),
-    assert(xmodule(Module, Src)).
+    assert(xmodule(Module, Src)),
+    (   module_property(Module, class(system))
+    ->  retractall(xoption(Src, register_called(_))),
+        assert(xoption(Src, register_called(all)))
+    ;   true
+    ).
 
 assert_module_export(_, []) :- !.
 assert_module_export(Src, [H|T]) :-
@@ -2290,7 +2382,6 @@ process_predicate_comma(A, Closure, Src) :-
     call(Closure, A, Src).
 
 
-assert_dynamic(_M:_Name/_Arity, _Src) :- !.   % not local
 assert_dynamic(PI, Src) :-
     pi_to_head(PI, Term),
     (   thread_local(Term, Src, _)  % dynamic after thread_local has
@@ -2299,7 +2390,6 @@ assert_dynamic(PI, Src) :-
         assert(dynamic(Term, Src, Line))
     ).
 
-assert_thread_local(_M:_Name/_Arity, _Src) :- !. % not local
 assert_thread_local(PI, Src) :-
     pi_to_head(PI, Term),
     current_source_line(Line),
