@@ -63,6 +63,7 @@
                        workers(positive_integer),
                        timeout(number),
                        keep_alive_timeout(number),
+                       silent(boolean),
                        ssl(list(any)),  % if http/http_ssl_plugin is loaded
                        pass_to(system:thread_create/3, 3)
                      ]).
@@ -156,6 +157,10 @@ self-signed SSL certificate.
 %     If you need to control resource usage look at the `spawn`
 %     option of http_handler/3 and library(thread_pool).
 %
+%     * silent(Bool)
+%     If `true` (default `false`), do not print an informational
+%     message that the server was started.
+%
 %   A  typical  initialization  for  an    HTTP   server  that  uses
 %   http_dispatch/1 to relay requests to predicates is:
 %
@@ -172,14 +177,17 @@ self-signed SSL certificate.
 %   and HTTPS server, where the HTTP   server redirects to the HTTPS
 %   server for handling sensitive requests.
 
-http_server(Goal, M:Options) :-
-    select_option(port(Port), Options, Options1),
+http_server(Goal, M:Options0) :-
+    option(port(Port), Options0),
     !,
-    make_socket(Port, M:Options1, Options2),
-    create_workers(Options2),
-    create_server(Goal, Port, Options2),
-    print_message(informational,
-                  httpd_started_server(Port)).
+    make_socket(Port, M:Options0, Options),
+    create_workers(Options),
+    create_server(Goal, Port, Options),
+    (   option(silent(true), Options0)
+    ->  true
+    ;   print_message(informational,
+                      httpd_started_server(Port))
+    ).
 http_server(_Goal, _Options) :-
     existence_error(option, port).
 
@@ -244,9 +252,11 @@ create_server(Goal, Address, Options) :-
     scheme(Scheme, Options),
     address_port(Address, Port),
     make_addr_atom(Scheme, Port, Alias),
-    thread_create(accept_server(Goal, Options), _,
+    thread_self(Initiator),
+    thread_create(accept_server(Goal, Initiator, Options), _,
                   [ alias(Alias)
                   ]),
+    thread_get_message(server_started),
     assert(current_server(Port, Goal, Alias, Queue, Scheme, StartTime)).
 
 scheme(Scheme, Options) :-
@@ -356,18 +366,19 @@ http_current_worker(Port, ThreadID) :-
     queue_worker(Queue, ThreadID).
 
 
-%!  accept_server(:Goal, +Options)
+%!  accept_server(:Goal, +Initiator, +Options)
 %
 %   The goal of a small server-thread accepting new requests and
 %   posting them to the queue of workers.
 
-accept_server(Goal, Options) :-
-    catch(accept_server2(Goal, Options), http_stop, true),
+accept_server(Goal, Initiator, Options) :-
+    catch(accept_server2(Goal, Initiator, Options), http_stop, true),
     thread_self(Thread),
     retract(current_server(_Port, _, Thread, _Queue, _Scheme, _StartTime)),
     close_server_socket(Options).
 
-accept_server2(Goal, Options) :-
+accept_server2(Goal, Initiator, Options) :-
+    thread_send_message(Initiator, server_started),
     repeat,
       (   catch(accept_server3(Goal, Options), E, true)
       ->  (   var(E)
@@ -569,7 +580,8 @@ http_worker(Options) :-
           thread_detach(Self),
           (   Sender == idle
           ->  true
-          ;   thread_send_message(Sender, quitted(Self))
+          ;   retract(queue_worker(Queue, Self)),
+              thread_send_message(Sender, quitted(Self))
           )
       ;   open_client(Message, Queue, Goal, In, Out,
                       Options, ClientOptions),
@@ -665,8 +677,9 @@ check_keep_alive_connection(In, TMO, Peer, In, Out) :-
 
 done_worker :-
     thread_self(Self),
-    thread_property(Self, status(Status)),
     retract(queue_worker(Queue, Self)),
+    thread_property(Self, status(Status)),
+    !,
     (   catch(recreate_worker(Status, Queue), _, fail)
     ->  thread_detach(Self),
         print_message(informational,
@@ -675,6 +688,12 @@ done_worker :-
         print_message(Level,
                       httpd_stopped_worker(Self, Status))
     ).
+done_worker :-                                  % received quit(Sender)
+    thread_self(Self),
+    thread_property(Self, status(Status)),
+    done_status_message_level(Status, Level),
+    print_message(Level,
+                  httpd_stopped_worker(Self, Status)).
 
 done_status_message_level(true, silent) :- !.
 done_status_message_level(exception('$aborted'), silent) :- !.
@@ -683,11 +702,18 @@ done_status_message_level(_, informational).
 
 %!  recreate_worker(+Status, +Queue) is semidet.
 %
-%   Deal with the possibility that  threads are, during development,
-%   killed with abort/0. We  recreate  the   worker  to  avoid  that
-%   eventually we run out of workers.  If   we  are aborted due to a
-%   halt/0 call, thread_create/3 will raise a permission error.
+%   Deal with the possibility  that   threads  are,  during development,
+%   killed with abort/0. We recreate the worker to avoid that eventually
+%   we run out of workers. If  we  are   aborted  due  to a halt/0 call,
+%   thread_create/3 will raise a permission error.
+%
+%   The first clause deals with the possibility  that we cannot write to
+%   `user_error`. This is possible when Prolog   is started as a service
+%   using some service managers. Would be  nice   if  we  could write an
+%   error, but where?
 
+recreate_worker(exception(error(io_error(write,user_error),_)), _Queue) :-
+    halt(2).
 recreate_worker(exception(Error), Queue) :-
     recreate_on_error(Error),
     queue_options(Queue, Options),
