@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2011-2016, University of Amsterdam
+    Copyright (c)  2011-2017, University of Amsterdam
                               VU University Amsterdam
     All rights reserved.
 
@@ -45,9 +45,6 @@
 #endif
 #include <time.h>
 
-#define LOCK()   PL_LOCK(L_PLFLAG)
-#define UNLOCK() PL_UNLOCK(L_PLFLAG)
-
 
 		 /*******************************
 		 *	PROLOG FLAG HANDLING	*
@@ -222,29 +219,58 @@ setPrologFlag(const char *name, int flags, ...)
 
 static void
 freePrologFlag(prolog_flag *f)
-{ if ( (f->flags & FT_MASK) == FT_TERM )
-    PL_erase(f->value.t);
+{ switch((f->flags & FT_MASK))
+  { case FT_TERM:
+      PL_erase(f->value.t);
+      break;
+    case FT_ATOM:
+      PL_unregister_atom(f->value.a);
+      break;
+    default:
+      ;
+  }
 
   freeHeap(f, sizeof(*f));
 }
 
 
 #ifdef O_PLMT
-static void
-copySymbolPrologFlagTable(void *name, void **value)
-{ prolog_flag *f = *value;
-  prolog_flag *copy = allocHeapOrHalt(sizeof(*copy));
+static prolog_flag *
+copy_prolog_flag(const prolog_flag *f)
+{ prolog_flag *copy = allocHeapOrHalt(sizeof(*copy));
 
   *copy = *f;
-  if ( (f->flags & FT_MASK) == FT_TERM )
-    copy->value.t = PL_duplicate_record(f->value.t);
-  *value = copy;
+  switch((f->flags & FT_MASK))
+  { case FT_TERM:
+      copy->value.t = PL_duplicate_record(f->value.t);
+      break;
+    case FT_ATOM:
+      PL_register_atom(copy->value.a);
+      break;
+    default:
+      ;
+  }
+
+  return copy;
+}
+
+
+static void
+copySymbolPrologFlagTable(void *name, void **value)
+{ atom_t key = (atom_t)name;
+  prolog_flag *f = *value;
+
+  PL_register_atom(key);
+  *value = copy_prolog_flag(f);
 }
 
 
 static void
 freeSymbolPrologFlagTable(void *name, void *value)
-{ freePrologFlag(value);
+{ atom_t key = (atom_t)name;
+
+  PL_unregister_atom(key);
+  freePrologFlag(value);
 }
 #endif
 
@@ -330,12 +356,14 @@ setUnknown(term_t value, atom_t a, Module m)
     }
 
     if ( !SYSTEM_MODE )
-      printMessage(ATOM_warning, PL_CHARS, "unknown_in_module_user");
+    { if ( !printMessage(ATOM_warning, PL_CHARS, "unknown_in_module_user") )
+	return FALSE;
+    }
   }
 
   m->flags = flags;
 
-  succeed;
+  return TRUE;
 }
 
 
@@ -557,11 +585,7 @@ set_prolog_flag_unlocked(term_t key, term_t value, int flags)
 
 #ifdef O_PLMT
     if ( GD->statistics.threads_created > 1 )
-    { prolog_flag *f2 = allocHeapOrHalt(sizeof(*f2));
-
-      *f2 = *f;
-      if ( (f2->flags & FT_MASK) == FT_TERM )
-	f2->value.t = PL_duplicate_record(f2->value.t);
+    { f = copy_prolog_flag(f);
 
       if ( !LD->prolog_flag.table )
       { LD->prolog_flag.table = newHTable(4);
@@ -570,10 +594,10 @@ set_prolog_flag_unlocked(term_t key, term_t value, int flags)
 	LD->prolog_flag.table->free_symbol = freeSymbolPrologFlagTable;
       }
 
-      addNewHTable(LD->prolog_flag.table, (void *)k, f2);
+      addNewHTable(LD->prolog_flag.table, (void *)k, f);
+      PL_register_atom(k);
       DEBUG(MSG_PROLOG_FLAG,
 	    Sdprintf("Localised Prolog flag %s\n", PL_atom_chars(k)));
-      f = f2;
     }
 #endif
   } else if ( !(flags & FF_NOCREATE) )	/* define new Prolog flag */
@@ -821,9 +845,9 @@ int
 set_prolog_flag(term_t key, term_t value, int flags)
 { int rc;
 
-  LOCK();
+  PL_LOCK(L_PLFLAG);
   rc = set_prolog_flag_unlocked(key, value, flags);
-  UNLOCK();
+  PL_UNLOCK(L_PLFLAG);
 
   return rc;
 }
@@ -1177,7 +1201,7 @@ pl_prolog_flag5(term_t key, term_t value,
   }
 
   fid = PL_open_foreign_frame();
-  LOCK();
+  PL_LOCK(L_PLFLAG);
   for(;;)
   { atom_t fn;
     prolog_flag *f;
@@ -1193,7 +1217,7 @@ pl_prolog_flag5(term_t key, term_t value,
 	   (!scope  || PL_unify_atom(scope, e->scope)) &&
 	   (!access || unify_prolog_flag_access(f, access)) &&
 	   (!type   || unify_prolog_flag_type(f, type)) )
-      { UNLOCK();
+      { PL_UNLOCK(L_PLFLAG);
 	ForeignRedoPtr(e);
       }
       if ( exception_term )
@@ -1210,7 +1234,7 @@ pl_prolog_flag5(term_t key, term_t value,
     } else
       break;
   }
-  UNLOCK();
+  PL_UNLOCK(L_PLFLAG);
 
   freeTableEnum(e->table_enum);
   freeHeap(e, sizeof(*e));
@@ -1305,8 +1329,11 @@ initPrologFlags(void)
 #ifdef O_PLMT
   setPrologFlag("threads",	FT_BOOL|FF_READONLY, TRUE, 0);
   setPrologFlag("system_thread_id", FT_INTEGER|FF_READONLY, 0, 0);
+  setPrologFlag("gc_thread",    FT_BOOL,
+		truePrologFlag(PLFLAG_GCTHREAD), PLFLAG_GCTHREAD);
 #else
   setPrologFlag("threads",	FT_BOOL|FF_READONLY, FALSE, 0);
+  setPrologFlag("gc_thread",    FT_BOOL|FF_READONLY, FALSE, PLFLAG_GCTHREAD);
 #endif
 #ifdef O_DDE
   setPrologFlag("dde", FT_BOOL|FF_READONLY, TRUE, 0);
@@ -1396,9 +1423,16 @@ initPrologFlags(void)
   setPrologFlag("kernel_compile_mode", FT_ATOM|FF_READONLY, "debug");
 #endif
 
-#if defined(__DATE__) && defined(__TIME__)
+#if defined(BUILD_TIME) && defined(BUILD_DATE)
+  setPrologFlag("compiled_at", FT_ATOM|FF_READONLY, BUILD_DATE ", " BUILD_TIME);
+#elif defined(__DATE__) && defined(__TIME__)
   setPrologFlag("compiled_at", FT_ATOM|FF_READONLY, __DATE__ ", " __TIME__);
 #endif
+  setPrologFlag("error_ambiguous_stream_pair", FT_BOOL, FALSE,
+		PLFLAG_ERROR_AMBIGUOUS_STREAM_PAIR);
+#ifdef O_MITIGATE_SPECTRE
+  setPrologFlag("mitigate_spectre", FT_BOOL, FALSE, PLFLAG_MITIGATE_SPECTRE);
+#endif
 
   setTZPrologFlag();
   setOSPrologFlags();
