@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2012-2016, University of Amsterdam
+    Copyright (c)  2012-2017, University of Amsterdam
                               VU University Amsterdam
     All rights reserved.
 
@@ -52,9 +52,6 @@ option  parsing,  initialisation  and  handling  of errors and warnings.
 #include <process.h>			/* getpid() */
 #endif
 
-#define LOCK()   PL_LOCK(L_INIT)
-#define UNLOCK() PL_UNLOCK(L_INIT)
-
 #if defined(_DEBUG) && defined(__WINDOWS__) && !defined(__MINGW32__)
 #include <crtdbg.h>
 #endif
@@ -76,7 +73,7 @@ static bool	vsysError(const char *fm, va_list args);
 			  } \
 			}
 #define	optionList(l)   { if (argc > 1) \
-			  { opt_append(l, store_string(argv[1])); \
+			  { opt_append(l, argv[1]); \
 			    argc--; argv++; \
 			  } else \
 			  { return -1; \
@@ -139,11 +136,11 @@ is_longopt(const char *optstring, const char *name)
 }
 
 
-static int
-opt_append(opt_list **l, char *s)
+int
+opt_append(opt_list **l, const char *s)
 { opt_list *n = allocHeapOrHalt(sizeof(*n));
 
-  n->opt_val = s;
+  n->opt_val = store_string(s);
   n->next = NULL;
 
   while(*l)
@@ -322,6 +319,7 @@ setupGNUEmacsInferiorMode()
   int val;
 
   if ( ((s = Getenv("EMACS", envbuf, sizeof(envbuf))) && s[0]) ||
+       ((s = Getenv("INSIDE_EMACS", envbuf, sizeof(envbuf))) && s[0]) ||
        ((s = Getenv("INFERIOR", envbuf, sizeof(envbuf))) && streq(s, "yes")) )
   { GET_LD
 
@@ -347,14 +345,14 @@ initPaths(int argc, const char **argv)
 	 !(symbols = DeRefLink(symbols, plp)) )
       symbols = argv[0];
 
-    DEBUG(2, Sdprintf("rc-module: %s\n", symbols));
+    DEBUG(MSG_INITIALISE, Sdprintf("rc-module: %s\n", symbols));
 
     systemDefaults.home	       = findHome(symbols, argc, argv);
 
 #ifdef __WINDOWS__			/* we want no module but the .EXE */
     GD->paths.module	       = store_string(symbols);
     symbols = findExecutable(NULL, plp, sizeof(plp));
-    DEBUG(2, Sdprintf("Executable: %s\n", symbols));
+    DEBUG(MSG_INITIALISE, Sdprintf("Executable: %s\n", symbols));
 #endif
     GD->paths.executable       = store_string(symbols);
     GD->options.systemInitFile = defaultSystemInitFile(argv[0]);
@@ -412,7 +410,6 @@ cleanupPaths(void)
   cleanupStringP(&systemDefaults.startup);
   cleanupStringP(&GD->options.systemInitFile);
   cleanupStringP(&GD->options.compileOut);
-  cleanupStringP(&GD->options.goal);
   cleanupStringP(&GD->options.topLevel);
   cleanupStringP(&GD->options.initFile);
   cleanupStringP(&GD->options.saveclass);
@@ -422,6 +419,7 @@ cleanupPaths(void)
 #endif
 
   cleanupOptListP(&GD->options.scriptFiles);
+  cleanupOptListP(&GD->options.goals);
 }
 
 
@@ -434,8 +432,8 @@ initDefaults(void)
   systemDefaults.global      = DEFGLOBAL;
   systemDefaults.trail       = DEFTRAIL;
   systemDefaults.table       = DEFTABLE;
-  systemDefaults.goal	     = "version";
-  systemDefaults.toplevel    = "prolog";
+  systemDefaults.goal	     = NULL;
+  systemDefaults.toplevel    = "default";
   systemDefaults.notty       = NOTTYCONTROL;
 
 #ifdef __WINDOWS__
@@ -445,13 +443,16 @@ initDefaults(void)
   GD->io_initialised	     = FALSE;
   GD->initialised	     = FALSE;
   GD->bootsession	     = FALSE;
+#ifdef SIG_ALERT
+  GD->signals.sig_alert      = SIG_ALERT;
+#endif
 
   if ( systemDefaults.notty )
     clearPrologFlagMask(PLFLAG_TTY_CONTROL);
   else
     setPrologFlagMask(PLFLAG_TTY_CONTROL);
 
-  setPrologFlagMask(PLFLAG_DEBUGINFO);
+  setPrologFlagMask(PLFLAG_DEBUGINFO|PLFLAG_GCTHREAD);
 }
 
 
@@ -467,43 +468,40 @@ This  file  is  parsed,  and  the    values  are  interpreted.  See  the
 if-then-else below for the defined values.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#define MAXVARNAME 256
-#define MAXVARVAL  1024
-
 static int
-getVarFromStream(IOSTREAM *s, char *name, char *value)
-{ char *q;
-  int l;
-  int c;
+getVarFromStream(IOSTREAM *s, tmp_buffer *name, tmp_buffer *value)
+{ int c;
 
 again:
-  for(l=MAXVARNAME, q=name; --l > 0; )
+  initBuffer(name);
+  initBuffer(value);
+
+  for(;;)
   { switch(c = Sgetc(s))
     { case EOF:
 	return FALSE;
       case '=':
-	*q = EOS;
+	addBuffer(name, EOS, char);
         goto do_value;
       case '\n':
+	discardBuffer(name);
 	goto again;
       default:
-	*q++ = c;
+	addBuffer(name, c, char);
     }
   }
-  return FALSE;
 
 do_value:
-  for(l=MAXVARVAL, q=value; --l > 0; )
+  for(;;)
   { switch(c = Sgetc(s))
     { case EOF:
       case '\n':
-	*q = EOS;
+	addBuffer(value, EOS, char);
         return TRUE;
       default:
-	*q++ = c;
+	addBuffer(value, c, char);
     }
   }
-  return FALSE;
 }
 
 
@@ -514,21 +512,27 @@ initDefaultOptions(void)
   GD->options.globalSize    = systemDefaults.global   K;
   GD->options.trailSize     = systemDefaults.trail    K;
   GD->options.tableSpace    = systemDefaults.table    K;
-  GD->options.goal	    = store_string(systemDefaults.goal);
   GD->options.topLevel      = store_string(systemDefaults.toplevel);
   GD->options.initFile      = store_string(systemDefaults.startup);
   GD->options.scriptFiles   = NULL;
   GD->options.saveclass	    = store_string("none");
 
+  if ( systemDefaults.goal )
+    opt_append(&GD->options.goals, systemDefaults.goal);
+
+
   if ( !GD->bootsession && GD->resourceDB )
   { IOSTREAM *op = SopenRC(GD->resourceDB, "$options", "$prolog", RC_RDONLY);
 
     if ( op )
-    { char name[MAXVARNAME];
-      char val[MAXVARVAL];
+    { tmp_buffer name;
+      tmp_buffer val;
 
-      while( getVarFromStream(op, name, val) )
-	set_pl_option(name, val);
+      while( getVarFromStream(op, &name, &val) )
+      { set_pl_option(baseBuffer(&name, char), baseBuffer(&val, char));
+	discardBuffer(&name);
+	discardBuffer(&val);
+      }
 
       Sclose(op);
     }
@@ -564,6 +568,12 @@ parseCommandLineOptions(int argc0, char **argv, int *compile)
 { GET_LD
   int argc = argc0;
 
+  DEBUG(MSG_INITIALISE,
+	{ Sdprintf("parseCommandLineOptions");
+	  for(int i=0; i<argc; i++)
+	    Sdprintf("%s ", argv[i]);
+	});
+
   for( ; argc > 0 && (argv[0][0] == '-' || argv[0][0] == '+'); argc--, argv++ )
   { char *s = &argv[0][1];
 
@@ -580,6 +590,7 @@ parseCommandLineOptions(int argc0, char **argv, int *compile)
       continue;
     } else if ( isoption(s, "nosignals") )
     { clearPrologFlagMask(PLFLAG_SIGNALS);
+      clearPrologFlagMask(PLFLAG_GCTHREAD);
       continue;
     } else if ( isoption(s, "nodebug") )
     { clearPrologFlagMask(PLFLAG_DEBUGINFO);
@@ -604,6 +615,14 @@ parseCommandLineOptions(int argc0, char **argv, int *compile)
 #endif
       } else if ( (optval=is_longopt(s, "traditional")) )
       { setTraditional();
+      } else if ( (optval=is_longopt(s, "sigalert")) )
+      { char *e;
+	long sig = strtol(optval, &e, 10);
+
+	if ( e > optval && *e == EOS && sig >= 0 && sig < 32 )
+	  GD->signals.sig_alert = sig;
+	else
+	  return -1;
       }
 
       continue;				/* don't handle --long=value */
@@ -631,7 +650,7 @@ parseCommandLineOptions(int argc0, char **argv, int *compile)
 	case 'l':
 	case 's':	optionList(&GD->options.scriptFiles);
 			break;
-	case 'g':	optionString(GD->options.goal);
+	case 'g':	optionList(&GD->options.goals);
 			break;
 	case 't':	optionString(GD->options.topLevel);
 			break;
@@ -812,7 +831,7 @@ PL_initialise(int argc, char **argv)
 
   { GET_LD
 #ifdef HAVE_SIGNAL
-  setPrologFlagMask(PLFLAG_SIGNALS);	/* default: handle signals */
+    setPrologFlagMask(PLFLAG_SIGNALS);	/* default: handle signals */
 #endif
 
   if ( !GD->resourceDB )
@@ -845,7 +864,7 @@ PL_initialise(int argc, char **argv)
       }
     }
 
-    DEBUG(1, if (GD->bootsession) Sdprintf("Boot session\n"););
+    DEBUG(MSG_INITIALISE, if (GD->bootsession) Sdprintf("Boot session\n"););
 
     if ( !GD->resourceDB )
     { if ( !(GD->resourceDB = openResourceDB(argc, argv)) )
@@ -911,8 +930,7 @@ PL_initialise(int argc, char **argv)
     if ( statefd )
     { GD->bootsession = TRUE;
       if ( !loadWicFromStream(statefd) )
-      { fail;
-      }
+	return FALSE;
       GD->bootsession = FALSE;
 
       Sclose(statefd);
@@ -929,7 +947,7 @@ PL_initialise(int argc, char **argv)
   GD->initialised = TRUE;
   registerForeignLicenses();
 
-  DEBUG(1, Sdprintf("Starting Prolog Part of initialisation\n"));
+  DEBUG(MSG_INITIALISE, Sdprintf("Starting Prolog Part of initialisation\n"));
 
   if ( compile )
   { int status = prologToplevel(PL_new_atom("$compile")) ? 0 : 1;
@@ -1008,12 +1026,20 @@ usage(void)
   return TRUE;
 }
 
+#ifndef PLVERSION_TAG
+#define PLVERSION_TAG ""
+#endif
+
 static int
-version()
-{ Sprintf("SWI-Prolog version %d.%d.%d for %s\n",
+version(void)
+{ const char *tag = PLVERSION_TAG;
+
+  if ( !tag ) tag = "";
+  Sprintf("SWI-Prolog version %d.%d.%d%s%s for %s\n",
 	  PLVERSION / 10000,
 	  (PLVERSION / 100) % 100,
 	  PLVERSION % 100,
+	  tag[0] ? "-" : "", tag,
 	  PLARCH);
 
   return TRUE;
@@ -1052,6 +1078,7 @@ runtime_vars(int format)
   char base[MAXPATHLEN];
 #endif
   char version[20];
+  char *tag = PLVERSION_TAG;
 
   if ( systemDefaults.home )
   {
@@ -1084,6 +1111,8 @@ runtime_vars(int format)
   printvar("PLSOPATH",	SO_PATH, format);
 #endif
   printvar("PLVERSION", version, format);
+  if ( tag[0] )
+    printvar("PLVERSIONTAG", tag, format);
 #if defined(HAVE_DLOPEN) || defined(HAVE_SHL_LOAD) || defined(EMULATE_DLOPEN)
   printvar("PLSHARED",	"yes", format);
 #else
@@ -1205,9 +1234,9 @@ cleanupProlog(int rval, int reclaim_memory)
     PlMessage("Exit status is %d", rval);
 #endif
 
-  LOCK();
+  PL_LOCK(L_INIT);
   if ( GD->cleaning != CLN_NORMAL )
-  { UNLOCK();
+  { PL_UNLOCK(L_INIT);
     return FALSE;
   }
 
@@ -1224,18 +1253,17 @@ cleanupProlog(int rval, int reclaim_memory)
   debugmode(FALSE, NULL);		/* avoid recursive tracing */
 
   if ( GD->initialised )
-  { DEBUG(5, Sdprintf("Running at_halt hooks\n"));
+  { DEBUG(MSG_CLEANUP, Sdprintf("Running at_halt hooks\n"));
 
     if ( LD->outofstack )
       emptyStacks();
 
     PL_set_prolog_flag("exit_status", PL_INTEGER, rval);
-    if ( !query_loop(PL_new_atom("$run_at_halt"), FALSE) &&
-	 rval == 0 &&
-	 !PL_exception(0) )
+    if ( query_loop(PL_new_atom("$run_at_halt"), FALSE) == FALSE &&
+	 rval == 0 )
     { if ( ++GD->halt_cancelled	< MAX_HALT_CANCELLED )
       { GD->cleaning = CLN_NORMAL;
-	UNLOCK();
+	PL_UNLOCK(L_INIT);
 	return FALSE;
       }
     }
@@ -1244,7 +1272,7 @@ cleanupProlog(int rval, int reclaim_memory)
     if ( !run_on_halt(&GD->os.on_halt_list, rval) && rval == 0 )
     { if ( ++GD->halt_cancelled	< MAX_HALT_CANCELLED )
       { GD->cleaning = CLN_NORMAL;
-	UNLOCK();
+	PL_UNLOCK(L_INIT);
 	return FALSE;
       }
     }
@@ -1324,7 +1352,7 @@ emergency:
     cleanupDebug();
   }
 
-  UNLOCK();				/* requires GD->thread.enabled */
+  PL_UNLOCK(L_INIT);				/* requires GD->thread.enabled */
 
   if ( reclaim_memory )
   { memset(&PL_global_data, 0, sizeof(PL_global_data));
@@ -1337,7 +1365,7 @@ emergency:
 
 int
 PL_cleanup(int rc)
-{ return cleanupProlog(rc, TRUE);
+{ return cleanupProlog(rc, rc == 0);
 }
 
 
@@ -1446,7 +1474,7 @@ action:
 
   switch(getSingleChar(Sinput, FALSE))
   { case EOF:
-      Sfprintf(Serror, "EOF: exit\n");
+      Sfprintf(Serror, "EOF: exit (status 3)\n");
     case 'e':
       PL_halt(3);
       break;
